;;;
;;; $Header: /home/gene/icfp2001/source/RCS/evolver.lisp,v 1.2 2001/07/29 05:22:13 gene Exp $
;;;

(defvar *evolver-pool-size* nil
  "The desired size of the gene pool.  We remember this size in a variable of
it's own in case any of the pool-mangling code fudges the actual pool's size.
By keeping the truly desired pool size, we can help minimize cumulative pool
size fudging errors.")

(defvar *evolver-pool* nil
  "A gene pool.  Each gene is an Organism.")

(defvar *evolver-generation* 0 "A count of the generations.")

(defun evolver-init (size)
  "Initialize the gene pool from *MEANING*, which is assumed to hold the
parsed input file.  The initialized gene poll will have SIZE organisms in it."
  (when (null *meaning*)
    (format t "~&evolver-init: warning: ")
    (format t "*meaning* should hold a parsed input file, but it is nil."))
  (setf *evolver-pool-size* size)
  (setf *evolver-pool* (make-array size))
  (dotimes (i size)
	   (setf (aref *evolver-pool* i) (make-random-organism)))
  (evolver-sort))

(defun evolver-sort ()
  (sort *evolver-pool* #'(lambda (a b)
			   (< (organism-fitness a) (organism-fitness b)))))

(defun evolve ()
  (incf *evolver-generation*)
  (when *debug* (format t "~&evolver generation ~A" *evolver-generation*))
  (let ((pool (make-array *evolver-pool-size*))
	(i 0))
    (dotimes (j (/ (- *evolver-pool-size* i) 2))
	     (setf (aref pool i)
		   (make-child-organism (aref *evolver-pool* 0)
					(aref *evolver-pool* j)))
	     (incf i))
    (dotimes (j (/ (- *evolver-pool-size* i) 2))
	     (setf (aref pool i)
		   (make-child-organism (aref *evolver-pool* 1)
					(aref *evolver-pool* j)))
	     (incf i))
    (dotimes (j (- *evolver-pool-size* i))
	     (setf (aref pool i)
		   (make-child-organism (aref *evolver-pool* 2)
					(aref *evolver-pool* j)))
	     (incf i))
    (setq *evolver-pool* pool))
  (evolver-sort))

;;; --- end of file ---
