;;;
;;; $Header: /home/gene/icfp2001/source/RCS/sml.lisp,v 1.4 2001/07/29 09:17:34 gene Exp $
;;;

(defconstant *whitespace* '(#\space #\tab #\newline #\linefeed))

(defvar *sml-lastchar* nil
  "The last character we printed, or nil if there was no such character or if
the last thing printed was a tag.")

(defun sml-tag-p (token)
  (stringp token))

(defun sml-begin-tag-p (token)
  (and (sml-tag-p token) (not (sml-end-tag-p token))))

(defun sml-end-tag-p (token)
  (and (sml-tag-p token) (eq (aref token 0) #\<) (eq (aref token 1) #\/)))

(defun sml-tag-meaning (tag)
  "Returns a string that has just the meaning of a tag, with the <> stripped."
  (subseq tag
	  (if (sml-begin-tag-p tag)
	      1
	    2)
	  (- (length tag) 1)))

(defun sml-parse (&optional (strm *standard-input*))
  (let ((stack (list (make-context)))
	lst
	(input (read strm)))
    (mapc #'(lambda (token)
	      (cond ((sml-begin-tag-p token)
		     (setq stack (cons (child-context (sml-tag-meaning token)
					(car stack))
				       stack)))
		    ((sml-end-tag-p token)
		     (setq stack (cdr stack)))
		    (t (setq lst (cons (make-attrchar token (car stack))
				       lst)))))
	    input)
    (nreverse lst)))

(defun ensure-context-pop (desired stack strm)
  "Adjust the stack of contexts using only pop operations.  Return a new
stack."
  (do ()
      ((or (endp stack) (context-equal desired (cdr (car stack))))
       stack)
      (when (not (equal (car (car stack)) ""))
	(setq *sml-lastchar* nil)
	(format strm "</~A>" (car (car stack))))
      (setq stack (cdr stack)))
  (when (endp stack)
    (setq stack (ensure-context-push desired stack strm)))
  stack)

;; The previous, recursive, implementation of ensure-context-pop.
;;  (cond ((endp stack) (ensure-context desired stack 'push strm))
;;	((context-equal desired (cdr (car stack))) stack)
;;	(t (if (not (equal (car (car stack)) ""))
;;	       (format strm "</~A>" (car (car stack))))
;;	   (ensure-context desired (cdr stack) 'pop strm))))

(defun ensure-context-push (desired stack strm)
  "Adjust the stack of contexts using only push operations.  Return a new
stack."
  (if (endp stack)
      (ensure-context desired (list (cons "" (make-context))) 'push strm)
    (let* ((top (car stack))
	   (tag (car top))
	   (ctx (cdr top)))
      (cond ((and (eq (context-size desired) nil)
		  (not (eq (context-size ctx) nil)))
	     (ensure-context-pop desired stack strm))
	    ((and (eq (context-color desired) nil)
		  (not (eq (context-color ctx) nil)))
	     (ensure-context-pop desired stack strm))
	    ((or (and (not (context-bold desired)) (context-bold ctx))
		 (and (not (context-italic desired)) (context-italic ctx))
		 (and (not (context-strong desired)) (context-strong desired))
		 (and (not (context-teletype desired))
		      (context-teletype desired))
		 (< (context-underline desired) (context-underline ctx)))
	     (setq *sml-lastchar* nil)
	     (format strm "<PL>")
	     (ensure-context desired
			     (acons "PL" (child-context "PL" ctx) stack)
			     'push
			     strm))
	    ((xor (context-emphasis desired) (context-emphasis ctx))
	     (setq *sml-lastchar* nil)
	     (format strm "<EM>")
	     (ensure-context desired
			     (acons "EM" (child-context "EM" ctx) stack)
			     'push
			     strm))
	    ((and (context-bold desired) (not (context-bold ctx)))
	     (setq *sml-lastchar* nil)
	     (format strm "<B>")
	     (ensure-context desired (acons "B" (child-context "B" ctx) stack)
			     'push strm))
	    ((< (context-underline ctx) (context-underline desired))
	     (setq *sml-lastchar* nil)
	     (format strm "<U>")
	     (ensure-context desired (acons "U" (child-context "U" ctx) stack)
			     'push strm))
	    ((and (context-italic desired) (not (context-italic ctx)))
	     (setq *sml-lastchar* nil)
	     (format strm "<I>")
	     (ensure-context desired (acons "I" (child-context "I" ctx) stack)
			     'push strm))
	    ((and (context-strong desired) (not (context-strong ctx)))
	     (setq *sml-lastchar* nil)
	     (format strm "<S>")
	     (ensure-context desired (acons "S" (child-context "S" ctx) stack)
			     'push strm))
	    ((and (context-teletype desired) (not (context-teletype ctx)))
	     (setq *sml-lastchar* nil)
	     (format strm "<TT>")
	     (ensure-context desired
			     (acons "TT" (child-context "TT" ctx) stack)
			     'push strm))
	    ((not (equal (context-size desired) (context-size ctx)))
	     (setq *sml-lastchar* nil)
	     (format strm "<~A>" (context-size desired))
	     (ensure-context desired
			     (acons (context-size desired)
				    (child-context (context-size desired) ctx)
				    stack)
			     'push
			     strm))
	    ((not (equal (context-color desired) (context-color ctx)))
	     (setq *sml-lastchar* nil)
	     (format strm "<~A>" (context-color desired))
	     (ensure-context desired
			     (acons (context-color desired)
				    (child-context (context-color desired) ctx)
				    stack)
			     'push
			     strm))))))

(defun supressible-p (attrchar lastchar)
  (when (and (whitespace-p (car attrchar))
       (whitespace-p (car lastchar))
       (eql (context-strong (cdr attrchar)) (context-strong (cdr lastchar)))
       (equal (context-size (cdr attrchar)) (context-size (cdr lastchar)))
       (or (equal (context-color (cdr attrchar))
		  (context-color (cdr lastchar)))
	   (equal (context-underline attrchar) 0)))
    (when *debug*
      (format t "~&supressible-p: debug: supressing ~A, lastchar is ~A."
	      attrchar lastchar))
    t))

(defun ensure-context (desired stack action strm)
  "Adjust the stack of (tag, context) frames until the top of stack's context
is equivalent to the desired context.  Action deterines the strategy used to
adjust the stack."
  (cond ((context-equal desired (cdr (car stack))) stack)
	((eq action 'push)  (ensure-context-push desired stack strm))
	((eq action 'pop)   (ensure-context-pop desired stack strm))))

(defun sml-print (attrchars actions &optional (strm *standard-output*))
  ;;
  ;; Each frame in stack is a CONS whose CAR is the tag that created
  ;; it & whose CDR is a context.
  ;;
  (setq *sml-lastchar* nil)
  (let (stack)
    (mapc #'(lambda (attrchar action)
	      (when (or (not (whitespace-p (car attrchar)))
			(context-teletype (cdr attrchar))
			(not (supressible-p attrchar *sml-lastchar*)))
		(setq stack (ensure-context (cdr attrchar) stack action strm))
		(format strm "~A" (car attrchar))
		(setq *sml-lastchar* attrchar)))
	  attrchars (action-ensure actions attrchars))
    (mapc #'(lambda (frame)
	      (unless (equal (car frame) "")
		     (format strm "</~A>" (car frame))))
	  stack))
  nil)


;;		(if (and (member (car attrchar) *whitespace*)
;;			 (not (context-teletype (cdr (car stack))))
;;			 (context-equal (cdr (car stack)) (cdr lastchar)))
;;		  (unless (member lastchar *whitespace*)
;;		    (format strm "~%"))
;;		(format strm "~A" (car attrchar)))
;;	      (setq lastchar attrchar))

;;; --- end of file ---
